from civipy.base.base import CiviCRMBase
from civipy.base.base import get_unique
from civipy.contact import CiviContact


class CiviContribution(CiviCRMBase):
    civicrm_entity_table = "contribution"

    def complete_transaction(self, **kwargs):
        """Calls the CiviCRM API completetransaction action and returns parsed JSON on success."""
        kwargs["id"] = self.civi_id
        response = self.action("completetransaction", **kwargs)
        if not isinstance(response.get("values"), int):
            value = get_unique(response)
            return self.__class__(value)
        else:
            return response

    @classmethod
    def find_by_transaction_id(cls, trxn_id: str, select: list[str] | None = None):
        """Find a contribution by payment transaction ID"""
        return cls.find(select=select, trxn_id=trxn_id)

    @classmethod
    def find_by_invoice_id(cls, invoice_id: str, select: list[str] | None = None):
        return cls.find(select=select, invoice_id=invoice_id)

    @classmethod
    def find_by_donor(
        cls,
        display_name: str,
        total_amount: float | None = None,
        receive_date: str | None = None,
        select: list[str] | None = None,
    ):
        """Find a contribution by donor's display name, and optionally
        by amount and/or date received (yyyy-mm-dd)."""
        contact = CiviContact.find(select=["contact_id"], display_name=display_name)
        return cls.find_by_donor_id(contact["contact_id"], total_amount, receive_date, select)

    @classmethod
    def find_by_donor_id(
        cls,
        contact_id: int,
        total_amount: float | None = None,
        receive_date: str | None = None,
        select: list[str] | None = None,
    ):
        """Find a contribution by donor's contact ID, and optionally
        by amount and/or date received (yyyy-mm-dd)."""
        query = {"contact_id": contact_id}
        if total_amount is not None:
            query["total_amount"] = total_amount
        if receive_date is not None:
            query["receive_date"] = {"BETWEEN": [receive_date, f"{receive_date} 23:59:59"]}
        return cls.find(select=select, **query)


class CiviContributionRecur(CiviCRMBase):
    civicrm_entity_table = "contributionrecur"

    @classmethod
    def find_by_transaction_id(cls, trxn_id: str, select: list[str] | None = None):
        """
        Find a recurring contribution by subscription transaction ID
        """
        return cls.find(select=select, trxn_id=trxn_id)



from civipy.base.base import CiviCRMBase
from civipy.activity import CiviActivity, CiviActivityContact
from civipy.address import CiviAddress, CiviLocationType, CiviStateProvince, CiviCountry
from civipy.contact import CiviContact, CiviEmail, CiviPhone, CiviRelationship, CiviTag, CiviEntityTag, CiviWebsite
from civipy.contact import CiviGroup, CiviGroupContact
from civipy.contribution import CiviContribution, CiviContributionRecur
from civipy.event import CiviEvent, CiviParticipant
from civipy.financial import CiviEntityFinancialTrxn
from civipy.grant import CiviGrant
from civipy.mailing import CiviMailingEventQueue
from civipy.membership import CiviMembership, CiviMembershipPayment
from civipy.note import CiviNote
from civipy.option_group import CiviCustomField, CiviCustomValue, CiviOptionGroup, CiviOptionValue
from civipy.order import CiviOrder, CiviPayment
from civipy.user import CiviUFField, CiviUFGroup, CiviUFJoin, CiviUFMatch, CiviUser

# enable e.g.
# >>> import civipy
# >>> civipy.init(rest_base="", ...)
from civipy.base.config import SETTINGS

init = SETTINGS.init
del SETTINGS
__all__ = [
    "CiviActivity",
    "CiviActivityContact",
    "CiviAddress",
    "CiviContact",
    "CiviContribution",
    "CiviContributionRecur",
    "CiviCountry",
    "CiviCustomField",
    "CiviCustomValue",
    "CiviEmail",
    "CiviEntityFinancialTrxn",
    "CiviEntityTag",
    "CiviEvent",
    "CiviGrant",
    "CiviGroup",
    "CiviGroupContact",
    "CiviLocationType",
    "CiviMailingEventQueue",
    "CiviMembership",
    "CiviMembershipPayment",
    "CiviNote",
    "CiviOptionGroup",
    "CiviOptionValue",
    "CiviOrder",
    "CiviParticipant",
    "CiviPayment",
    "CiviPhone",
    "CiviRelationship",
    "CiviStateProvince",
    "CiviTag",
    "CiviUFField",
    "CiviUFGroup",
    "CiviUFJoin",
    "CiviUFMatch",
    "CiviUser",
    "CiviWebsite",
    "CiviCRMBase",
    "init",
]



from civipy import CiviContribution


def test_find_with_existing():
    contrib = CiviContribution.find_by_transaction_id(trxn_id="12345")

    assert isinstance(contrib, CiviContribution)
    assert contrib.display_name == "Sample Donor"
    assert contrib.trxn_id == "12345"
    assert contrib.civi["display_name"] == "Sample Donor"
    assert contrib.civi["trxn_id"] == "12345"

    # do some tests of setattr while we have an object
    #
    # existing civi attributes can be modified
    contrib.display_name = "New Name"
    assert contrib.civi["display_name"] == "New Name"
    assert "display_name" not in contrib.__dict__
    assert contrib.display_name == "New Name"

    # but other attributes won't affect civi dict
    contrib.foo = "bar"
    assert "foo" not in contrib.civi
    assert contrib.foo == "bar"

    # unless they are prefixed with civi_
    contrib.civi_foo = "bar"
    assert contrib.civi["foo"] == "bar"
    assert contrib.foo == "bar"



import logging
import os
import pathlib
import sys
from typing import Callable, Literal
from civipy.exceptions import CiviProgrammingError

try:
    if sys.version_info >= (3, 11):
        import tomllib
    else:
        import tomli as tomllib
except ImportError:
    tomllib = None


class Settings:
    def __init__(self):
        self.values: dict[str, Setting] = {
            "rest_base": Setting(required=True),
            "user_key": Setting(required=True, private=True),
            "site_key": Setting(required=lambda: self.values["api_version"].value == "3", private=True),
            "log_file": Setting(),
            "log_level": Setting(default="DEBUG"),
            "api_version": Setting(default="4"),
        }
        self._api_type: Literal["http", "drush", "wp", "cvcli"] | None = None

    def init(self, **kwargs) -> None:
        """set configuration settings then set up logging"""
        for key, value in kwargs.items():
            if key not in self.values:
                continue
            self.values[key].value = value

        # set unset values to default or raise error if they are required
        missing = []
        for key, setting in self.values.items():
            if setting.required and (setting.not_set or not setting.value):
                missing.append(key)
            elif setting.not_set:
                self.values[key].value = None
        if missing:
            raise CiviProgrammingError(f"Missing required configuration values: {missing}")

        # set up logging - if log_file was not set, it will be None and basicConfig will ignore it and set up stream
        logging.basicConfig(filename=self.values["log_file"].value, level=self.values["log_level"].value)

    @property
    def api_type(self) -> str:
        if self._api_type is None:
            self._read_config()
            # determine value for api_type
            base = self.rest_base.lower()
            if "http" in base:
                self._api_type = "http"
            elif "drush" in base:
                self._api_type = "drush"
            elif "wp" in base:
                self._api_type = "wp"
            else:
                self._api_type = "cvcli"
        return self._api_type

    def _read_config(self) -> None:
        """Read configuration from wherever found"""
        values, source = {}, None
        # Read .civipy or pyproject.toml's tool.civipy section.
        file = find_config_file()
        if file is not None:
            if file.name == "pyproject.toml":
                values, source = read_pyproject_toml(file), str(file)
            else:
                values, source = read_dot_civipy(file), str(file)
        # Read any settings set by environment variables
        env_values = {key[5:].lower(): value for key, value in os.environ.items() if key.startswith("CIVI_")}
        if env_values:
            source = " and ".join(filter(None, (source, "environment variables")))
            values.update(env_values)
        self.init(**values)
        logger.debug("CiviPy settings read from %s", source)

    def __getattr__(self, item: str):
        """Look up config value, with fallback to read environment variables or config files"""
        if item not in self.values:
            return super().__getattribute__(item)
        if self.values[item].not_set:
            self._read_config()
        return self.values[item].value

    def __repr__(self):
        values = []
        for k, v in self.values.items():
            if v.not_set:
                value = "[UNSET]"
            elif v.value is None:
                value = "None"
            elif v.private:
                value = "[SET]"
            else:
                value = f"'{v.value}'"
            values.append(f"{k}={value}")
        return f"<Settings {' '.join(values)}>"


class Setting:
    def __init__(self, default: str | None = None, required: bool | Callable[[], bool] = False, private: bool = False):
        self.not_set = True
        self.private = private
        self._is_required = required
        self._default = default
        self._value: str | None = None

    @property
    def required(self) -> bool:
        return self._is_required() if callable(self._is_required) else self._is_required

    @property
    def value(self) -> str | None:
        return None if self.not_set else self._value

    @value.setter
    def value(self, new_value: str | None) -> None:
        self.not_set = False
        self._value = self._default if new_value is None else new_value

    def __repr__(self):
        return self.value or ""


def read_dot_civipy(file: pathlib.Path) -> dict[str, str | None]:
    """read settings from .civipy file"""
    values = {}
    with file.open() as fp:
        for line in fp:
            if line.startswith("#"):
                continue
            key, _, line_value = line.partition("=")
            value, _, comment = line_value.partition(" #")
            values[key.lower().strip()] = value.strip() or None
    return values


def read_pyproject_toml(file: pathlib.Path) -> dict[str, str | None]:
    """read settings from pyproject.toml file"""
    if tomllib is None:
        return {}
    with open(file, "rb") as fp:
        pyproject_toml = tomllib.load(fp)
    config = pyproject_toml.get("tool", {}).get("civipy", {})
    return {key.replace("-", "_"): value or None for key, value in config.items()}


def find_config_file() -> pathlib.Path | None:
    """Find a .civipy or pyproject.toml file. Look in current directory up to
    project root then check user's home directory for .civipy."""
    env_config = os.environ.get("CIVI_CONFIG")
    if env_config:
        directory = pathlib.Path(env_config).resolve()
        if directory.is_file():
            return directory
        dot_civipy = directory / ".civipy"
        if directory.is_dir() and dot_civipy.is_file():
            return dot_civipy
    start = pathlib.Path.cwd().resolve()
    for directory in [start] + list(start.parents):
        dot_civipy = directory / ".civipy"
        if dot_civipy.is_file():
            return dot_civipy
        pyproject = directory / "pyproject.toml"
        if pyproject.is_file():
            return pyproject
        if (directory / ".git").exists():
            break  # a .git folder suggests this is the project root
    dot_civipy = pathlib.Path.home().resolve() / ".civipy"
    if dot_civipy.is_file():
        return dot_civipy
    return None


SETTINGS = Settings()
logger = logging.getLogger("civipy")
__all__ = ["SETTINGS", "logger"]



from civipy.base.base import CiviCRMBase
from civipy.exceptions import NonUniqueResultError


class CiviMailingEventQueue(CiviCRMBase):
    """This is the table that links CiviMail mailings to Contacts.

    Attributes:
        id: str e.g. "1202"
        job_id: str e.g. "27"
        email_id: str e.g. "96173"
        hash: str e.g. "1fc2d50bfafd5c27"
        contact_id: str e.g. "16257"
    """

    @classmethod
    def find_recipients(cls, contact_ids: list[int]) -> "CiviMailingEventQueue | None":
        result = []
        for contact_id in set(contact_ids):
            found = cls.find_all(contact_id=contact_id)
            if not found:
                continue
            result.append(found[0])
        if not result:
            return None
        if len(result) != 1:
            raise NonUniqueResultError("Too many results")
        result = result[0]
        for attr in ("id", "job_id", "email_id", "contact_id"):
            if attr in result.civi:
                result.civi[attr] = int(result.civi[attr])
        return result



import json
from typing import Iterable, Literal
from civipy.base.base import CiviCRMBase, CiviEntity
from civipy.address import CiviCountry
from civipy.exceptions import CiviProgrammingError
from civipy.interface import CiviValue
from civipy.note import CiviNotable


class CiviContact(CiviNotable):
    civicrm_entity_table = "contact"

    @classmethod
    def find_by_email(cls, email_address: str, select: list[str] | None = None):
        if select is None:
            select = []
        email_obj = CiviEmail.objects.filter(email=email_address).values("contact_id")[0]
        return cls.objects.filter(id=email_obj.civi["contact_id"]).values(*select)[0]

    @classmethod
    def find_all_by_email(cls, email_address: str, select: list[str] | None = None):
        if select is None:
            select = []
        return [
            cls.objects.filter(id=email_obj.civi["contact_id"]).values(*select)[0]
            for email_obj in CiviEmail.objects.filter(email=email_address).values("contact_id").all()
        ]

    @classmethod
    def create(cls, **kwargs: CiviValue) -> "CiviContact":
        country_code: str | None = kwargs.pop("country_code", None)
        if country_code is not None:
            country = CiviCountry.find_by_country_code(country_code=country_code, select=["id"])
            kwargs["country_id"] = country.id
        return super().create(**kwargs)


class CiviEmail(CiviCRMBase):
    pass


class CiviPhone(CiviCRMBase):
    pass


class CiviWebsite(CiviCRMBase):
    pass


class CiviRelationship(CiviCRMBase):
    @classmethod
    def find_all(cls, select: list[str] | None = None, **kwargs: CiviValue) -> list[CiviEntity]:
        if select is None:
            return cls.objects.filter(**kwargs).all()
        return cls.objects.filter(**kwargs).values(*select).all()

    @classmethod
    def query_filter_hook(cls, version: Literal["3", "4"], query: CiviValue) -> CiviValue:
        if version == "3" or "where" not in query:
            return query
        contact_id = None
        parts = []
        for k, c, v in query["where"]:
            if k == "contact_id":
                contact_id = v
                parts.append(["OR", [["contact_id_a", "=", contact_id], ["contact_id_b", "=", contact_id]]])
                continue
            parts.append([k, c, v])
        if contact_id is not None:
            query["where"] = parts
        return query

    @classmethod
    def create_or_increment_relationship(  # noqa PLR0913 (too many args)
        cls,
        contact_id_a: int,
        contact_id_b: int,
        relationship_type_id: int,
        event_id: int | None = None,
        activity_id: int | None = None,
    ):
        print(f"in create_or_increment_relationship with a {contact_id_a} b {contact_id_b} type {relationship_type_id}")
        if not event_id and not activity_id:
            raise CiviProgrammingError("Must provide either event_id or activity_id")

        existing_relationship = cls.find(
            contact_id_a=contact_id_a,
            contact_id_b=contact_id_b,
            relationship_type_id=relationship_type_id,
        )
        if existing_relationship is None:
            # look for reverse relationship
            existing_relationship = cls.find(
                contact_id_a=contact_id_b,
                contact_id_b=contact_id_a,
                relationship_type_id=relationship_type_id,
            )

        events = [event_id] if event_id else []
        activities = [activity_id] if activity_id else []
        if existing_relationship is None:
            # create new relationship
            cls.create(
                contact_id_a=contact_id_a,
                contact_id_b=contact_id_b,
                relationship_type_id=relationship_type_id,
                description=json.dumps({"events": events, "activities": activities}),
                debug=1,
            )
        else:
            # update existing relationship
            relationship_info = {}
            if "description" in existing_relationship.civi:
                desc = existing_relationship.civi_description
                try:
                    relationship_info = json.loads(desc)
                except json.decoder.JSONDecodeError:
                    pass

            if "events" not in relationship_info:
                events.extend(relationship_info["events"])
            relationship_info["events"] = list(set(filter(None, events)))

            if "activities" not in relationship_info:
                activities.extend(relationship_info["activities"])
            relationship_info["activities"] = list(set(filter(None, activities)))

            existing_relationship.update(description=json.dumps(relationship_info))


class CiviEntityTag(CiviCRMBase):
    pass


class CiviTag(CiviCRMBase):
    pass


class CiviGroupContact(CiviCRMBase):
    pass


class CiviGroup(CiviCRMBase):
    @classmethod
    def find_by_title(cls, title: str, select: list[str] | None = None) -> "CiviGroup":
        """Finds a CiviGroup object for the group entitled "title"."""
        return cls.find(select=select, title=title)

    def update_all(self, contacts: Iterable[int | CiviContact]) -> None:
        query = CiviGroupContact.objects.filter(group_id=self.id).values("contact_id")
        current_pop = {c.contact_id for c in query}
        contact_ids = {c.id if isinstance(c, CiviContact) else c for c in contacts}
        for contact_id in current_pop.difference(contact_ids):
            CiviGroupContact.objects.filter(group_id=self.id, contact_id=contact_id).delete()
        for contact_id in contact_ids.difference(current_pop):
            CiviGroupContact.objects.create(group_id=self.id, contact_id=contact_id)

    def add_member(self, civi_contact: CiviContact) -> CiviGroupContact:
        return CiviGroupContact.find_or_create(where={"contact_id": civi_contact.id, "group_id": self.id})



from civipy import CiviContact


def test_get_with_existing():
    contact_info = CiviContact.objects.filter(email="validunique@example.com").all()

    assert len(contact_info) == 1
    assert isinstance(contact_info[0], CiviContact)


def test_get_no_match():
    contact_info = CiviContact.action("get", email="unknown@example.com")

    assert isinstance(contact_info, dict)
    assert contact_info["count"] == 0
    assert len(contact_info["values"]) == 0


def test_find_with_existing():
    contact = CiviContact.find(email="validunique@example.com")

    assert isinstance(contact, CiviContact)
    assert contact.display_name == "Valid Unique"
    assert contact.email == "validunique@example.com"
    assert contact.civi["display_name"] == "Valid Unique"
    assert contact.civi["email"] == "validunique@example.com"

    # do some tests of setattr while we have an object
    #
    # existing civi attributes can be modified
    contact.display_name = "New Name"
    assert contact.civi["display_name"] == "New Name"
    assert "display_name" not in contact.__dict__
    assert contact.display_name == "New Name"

    # but other attributes won't affect civi dict
    contact.foo = "bar"
    assert "foo" not in contact.civi
    assert contact.foo == "bar"

    # unless they are prefixed with civi_
    contact.civi_foo = "bar"
    assert contact.civi["foo"] == "bar"
    assert contact.foo == "bar"


def test_find_and_update_with_existing():
    contact = CiviContact.find_and_update(where={"email": "validunique@example.com"}, display_name="Updated Name")

    assert isinstance(contact, CiviContact)
    assert contact.display_name == "Updated Name"
    assert contact.email == "validunique@example.com"
    assert contact.civi["display_name"] == "Updated Name"
    assert contact.civi["email"] == "validunique@example.com"


def test_find_no_match():
    contact = CiviContact.find(email="unknown@example.com")

    assert contact is None


def test_find_all_no_match():
    contact = CiviContact.find_all(email="unknown@example.com")

    assert contact == []


def test_find_or_create_with_existing():
    contact = CiviContact.find_or_create(where={"email": "validunique@example.com"})

    assert isinstance(contact, CiviContact)
    assert contact.display_name == "Valid Unique"
    assert contact.email == "validunique@example.com"
    assert contact.civi["display_name"] == "Valid Unique"
    assert contact.civi["email"] == "validunique@example.com"



from typing import TypedDict, Literal, Callable
from warnings import warn
import urllib3

CiviValue = dict[str, int | str]
CiviV3Request = CiviValue


class CiviV3ResponseOptional(TypedDict, total=False):
    id: int
    error_message: str
    undefined_fields: list[str]


class CiviV3Response(CiviV3ResponseOptional):
    is_error: Literal[0, 1]
    version: Literal[3]
    count: int
    values: list[CiviValue] | dict[str, CiviValue] | CiviValue


class CiviV4Request(TypedDict, total=False):
    select: list[str]
    join: list[list[str | list[str]]]
    translationMode: Literal["fuzzy", "strict"]
    where: list[list[str | int]]
    values: list[CiviValue]
    orderBy: dict[str, Literal["ASC", "DESC"]]
    limit: int
    offset: int
    language: str
    chain: dict[str, list["str | CiviV4Request"]]
    groupBy: list[str]
    having: list[list[str | int]]
    debug: bool


class CiviV4ResponseOptional(TypedDict, total=False):
    error_code: int
    error_message: str
    debug: dict[str, str | list[str]]


class CiviV4Response(CiviV4ResponseOptional):
    entity: str
    action: str
    version: Literal[4]
    count: int
    countFetched: int
    values: list[CiviValue]


CiviResponse = CiviV3Response | CiviV4Response


class BaseInterface:
    """Interface manages communication with the CiviCRM API

    Call the Interface instance with `action`, `entity`, and `params` values to submit an API request.

    The `search_query` method is a helper function to generate query parameters.
    The `values` method is a helper function to generate parameters for create/update.
    """

    api_version: str = ""

    def __init__(self):
        self.func: Callable[[str, str, CiviValue], CiviResponse] | None = None
        self.http: urllib3.PoolManager | None = None

    def _configure_http_connection(self) -> None:
        timeout = urllib3.util.Timeout(connect=10.0, read=30.0)
        self.http = urllib3.PoolManager(timeout=timeout)

    def __call__(self, action: str, entity: str, params: CiviValue) -> CiviResponse:
        warn(
            "interface.__call__ will be removed in v0.1.0, use interface.execute instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.execute(action, entity, params)

    def execute(self, action: str, entity: str, params: CiviValue) -> CiviResponse:
        raise NotImplementedError

    operators = {
        "lte": "<=",
        "gte": ">=",
        "lt": "<",
        "gt": ">",
        "ne": "!=",
        "like": "LIKE",
        "not_like": "NOT LIKE",
        "in": "IN",
        "not_in": "NOT IN",
        "between": "BETWEEN",
        "not_between": "NOT BETWEEN",
        "isnull": "IS NULL",
    }

    @staticmethod
    def select(fields: list[str]) -> CiviValue | CiviV4Request:
        raise NotImplementedError

    @staticmethod
    def join(tables: list[tuple[str, str, str]]) -> CiviValue | CiviV4Request:
        raise NotImplementedError

    @staticmethod
    def sort(kwargs: CiviValue) -> CiviValue | CiviV4Request:
        raise NotImplementedError

    @staticmethod
    def limit(value: int, offset: int | None = None) -> CiviValue | CiviV4Request:
        raise NotImplementedError

    @staticmethod
    def where(kwargs: CiviValue) -> CiviValue | CiviV4Request:
        raise NotImplementedError

    @staticmethod
    def values(kwargs: CiviValue) -> CiviValue | CiviV4Request:
        raise NotImplementedError


__all__ = ["BaseInterface", "CiviValue", "CiviV3Response", "CiviV4Request", "CiviV4Response", "CiviResponse"]



from civipy.base.base import CiviCRMBase


class CiviNote(CiviCRMBase):
    pass


class CiviNotable(CiviCRMBase):
    def _where_for_note(self, subject: str) -> dict[str, str | int]:
        return {"entity_id": self.civi["id"], "entity_table": self.civicrm_entity_table, "subject": subject}

    def add_note(self, subject: str, note: str):
        return CiviNote.create(note=note, **self._where_for_note(subject))

    def find_or_create_note(self, subject: str, note: str):
        return CiviNote.find_or_create(where=self._where_for_note(subject), note=note)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from urllib3 import BaseHTTPResponse


class CivipyException(Exception):
    pass


class CiviProgrammingError(CivipyException):
    pass


class NoResultError(CivipyException):
    pass


class NonUniqueResultError(CivipyException):
    pass


class CiviHTTPError(CivipyException):
    def __init__(self, r: "BaseHTTPResponse"):
        self.r = r
        self.status_code = r.status
        self.message = r.data.decode("utf-8", errors="replace")
        super().__init__(f"{self.status_code}: {self.message}")


class CiviAPIError(CivipyException):
    def __init__(self, data):
        self.code = data.get("error_code")
        self.message = data.get("error_message")



from civipy.base.base import CiviCRMBase
from civipy.base.config import logger


class CiviCountry(CiviCRMBase):
    civicrm_entity_table = "country"

    @classmethod
    def find_by_country_code(cls, country_code: str, select: list[str] | None = None):
        return cls.find(select=select, iso_code=country_code)


class CiviStateProvince(CiviCRMBase):
    pass


class CiviLocationType(CiviCRMBase):
    pass


class CiviAddress(CiviCRMBase):
    country_iso_code_cache = {}
    state_province_abbreviation_cache = {}

    @classmethod
    def create(cls, **kwargs):
        """Calls the CiviCRM API create action and returns parsed JSON on success.

        Accepts field country_iso_code and will convert this to country_id
        Accepts field state_province_abbreviation and will convert this to state_province_id"""
        if "country_iso_code" in kwargs:
            country_iso_code = kwargs["country_iso_code"]
            del kwargs["country_iso_code"]
            if country_iso_code in cls.country_iso_code_cache:
                country = cls.country_iso_code_cache[country_iso_code]
            else:
                country = CiviCountry.find_by_country_code(country_iso_code)
                cls.country_iso_code_cache[country_iso_code] = country

            kwargs["country_id"] = country.civi_id

            if country.civi_id not in cls.state_province_abbreviation_cache:
                cls.state_province_abbreviation_cache[country.civi_id] = {}

            if "state_province_abbreviation" in kwargs:
                abbreviation = kwargs["state_province_abbreviation"]
                del kwargs["state_province_abbreviation"]
                if country.is_province_abbreviated:
                    if abbreviation in cls.state_province_abbreviation_cache[country.civi_id]:
                        logger.debug("Using cache for %s" % abbreviation)
                        state_province_id = cls.state_province_abbreviation_cache[country.civi_id][abbreviation]
                    else:
                        state_province = CiviStateProvince.find(country_id=country.civi_id, abbreviation=abbreviation)
                        state_province_id = state_province["id"]
                        cls.state_province_abbreviation_cache[country.civi_id][abbreviation] = state_province_id
                    kwargs["state_province_id"] = state_province_id
                else:
                    kwargs["supplemental_address_3"] = abbreviation
        return super().create(**kwargs)



from datetime import datetime
from typing import Literal
from civipy.base.base import CiviCRMBase
from civipy.contribution import CiviContribution
from civipy.exceptions import CiviProgrammingError
from civipy.interface import CiviValue


class CiviMembershipPayment(CiviCRMBase):
    pass


class CiviMembership(CiviCRMBase):
    def payments(self):
        """Find all MembershipPayment records associated with this Membership."""
        return CiviMembershipPayment.find_all(membership_id=self.id)

    def apply_contribution(self, contribution: CiviContribution):
        """Apply a Contribution to this Membership and extend the expiration date."""
        # The new expiration date should be the old expiration date plus one year.
        end = datetime.strptime(self.end_date, "%Y-%m-%d")
        end = end.replace(year=end.year + 1)
        # If the new expiration date is in the future, change the status to "Current".
        status = "2" if end > datetime.now() else self.status_id
        cid = contribution.id
        CiviMembershipPayment.create(contribution_id=cid, membership_id=self.id)
        self.update(end_date=end.strftime("%Y-%m-%d"), status_id=status)

    def set_status(self, status_id: int | None = None, status: str | None = None, is_override: bool = False):
        if status_id is not None and status is not None:
            raise CiviProgrammingError("Undefined behavior: called set_status with `status_id` and `status`.")
        if status_id is None and status is None:
            raise CiviProgrammingError("Called set_status with no status.")
        values = {"status_id": status_id}
        if is_override is True:
            values["is_override"] = True
        return self.objects.filter(id=self.id).update(**values)

    @classmethod
    def query_values_hook(cls, version: Literal["3", "4"], query: CiviValue) -> CiviValue:
        if version == "4" and "values" in query and isinstance(query["values"].get("status_id"), str):
            query["values"]["status_id.name"] = query["values"].pop("status_id")
        return query



from typing import TYPE_CHECKING
from civipy.exceptions import NoResultError, NonUniqueResultError, CiviProgrammingError

if TYPE_CHECKING:
    from civipy.interface import CiviValue, CiviResponse


def get_unique(response: "CiviResponse") -> "CiviValue":
    """Enforce that exactly one record was returned in `response` and return just the record."""
    if not response or "count" not in response or "values" not in response:
        raise CiviProgrammingError(f"Unexpected response {response}")
    record_id = next(iter(response["values"].keys())) if isinstance(response["values"], dict) else 0

    # Raise exception if `response` does not have exactly one record.
    count = response["count"]
    if count == 0:
        raise NoResultError("No results in response.")
    if count > 1 and not is_identical(response["values"]):
        raise NonUniqueResultError(f"Response is not unique, has {count} results.")

    return response["values"][record_id]


def is_identical(values: list["CiviValue"] | dict[str, "CiviValue"]) -> bool:
    """Check if multiple returned values are the same item and can be treated as unique."""
    # I have experienced the V4 REST API returning two instances of a record. The request was
    # `CiviContact.find(select=["*", "email_primary.email"], id=11722)`. This was likely due
    # to the join to the email table, so I'm treating this as an implicit "select distinct"
    if isinstance(values, dict):
        values = list(values.values())
    first = values[0]
    return all((value == first for value in values[1:]))



from typing import TYPE_CHECKING, Type, Optional
from civipy.interface import get_interface, Interface

if TYPE_CHECKING:
    from civipy.base.base import CiviCRMBase
    from civipy.interface import CiviValue


class BaseQuery:
    def __init__(self, model: Type["CiviCRMBase"]) -> None:
        self.model = model
        if model.__name__ in ("MetaCiviCRM", "CiviCRMBase"):
            self._entity: str | None = None
        else:
            self._entity = model.__name__[4:] if model.__name__[:4] == "Civi" else model.__name__
        self._select = None
        self._filter = None
        self._join = None
        self._limit = None
        self._order = None
        self._result_cache: list | None = None

    def __repr__(self) -> str:
        data = list(self[:21])
        if len(data) > 20:
            data[-1] = "...(remaining elements truncated)..."
        return f"<{self.__class__.__name__} {data}>"

    def __len__(self):
        self._fetch_all()
        return len(self._result_cache)

    def __iter__(self):
        self._fetch_all()
        return iter(self._result_cache)

    def __bool__(self):
        self._fetch_all()
        return bool(self._result_cache)

    def __getitem__(self, k: int | slice):
        """Retrieve an item or slice from the set of results."""
        if isinstance(k, int):
            if k < 0:
                raise ValueError("Negative indexing is not supported.")
            if self._result_cache is not None:
                return self._result_cache[k]
            query = self._chain()
            query._limit = {"offset": k, "limit": 1}
            query._fetch_all()
            return query._result_cache[0]

        if not isinstance(k, slice):
            raise TypeError(f"QuerySet indices must be integers or slices, not {type(k).__name__}.")
        if (k.start is not None and k.start < 0) or (k.stop is not None and k.stop < 0):
            raise ValueError("Negative indexing is not supported.")
        if self._result_cache is not None:
            return self._result_cache[k]
        query = self._chain()
        start, stop, step = [None if v is None else int(v) for v in (k.start, k.stop, k.step)]
        query._limit = {} if start is None else {"offset": start}
        query._limit["limit"] = stop - query._limit.get("offset", 0)
        return list(query)[::step] if step else query

    _interface_reference: Interface | None = None

    @classmethod
    def _interface(cls) -> Interface:
        """Instantiate the appropriate API interface and store it on the base Query class,
        so that it will be instantiated only once and available to all entity classes."""
        if BaseQuery._interface_reference is None:
            BaseQuery._interface_reference = get_interface()
        return BaseQuery._interface_reference

    def _chain(self):
        query = self.__class__(model=self.model)
        query._select = self._select
        query._filter = self._filter
        query._join = self._join
        query._limit = self._limit
        query._order = self._order
        return query

    def _compose(self, values: Optional["CiviValue"] = None) -> "CiviValue":
        """Create a query for the selected API version from the current attributes."""
        interface = self._interface()
        params = {}
        if self._select is not None:
            self._deep_update(params, interface.select(self._select))
        if self._join is not None:
            self._deep_update(params, interface.join(self._join))
        if self._filter is not None:
            where = interface.where(self._filter)
            if hasattr(self.model, "query_filter_hook"):
                where = self.model.query_filter_hook(interface.api_version, where)
            self._deep_update(params, where)
        if self._limit is not None:
            self._deep_update(params, interface.limit(self._limit["limit"]))
        if self._order is not None:
            self._deep_update(params, interface.sort(self._order))
        if values is not None:
            values = interface.values(values)
            if hasattr(self.model, "query_values_hook"):
                values = self.model.query_values_hook(interface.api_version, values)
            self._deep_update(params, values)
        return params

    def _deep_update(self, query: "CiviValue", other: "CiviValue") -> None:
        """Update `query` with the values in `other`, also updating contained dicts rather than overwriting them."""
        for key, val in other.items():
            if key in query and isinstance(query[key], dict) and isinstance(val, dict):
                self._deep_update(query[key], val)
                continue
            query[key] = val

    def _fetch_all(self):
        if self._result_cache is None:
            query = self._compose()
            results = self._interface().execute("get", self._entity, query)
            self._result_cache = list(map(self.model, results["values"]))


class Query(BaseQuery):
    def all(self):
        self._fetch_all()
        return self._result_cache

    def get(self, **kwargs):
        return self.filter(**kwargs)[0]

    def filter(self, **kwargs):
        query = self._chain()
        if query._filter:
            query._filter.update(kwargs)
        else:
            query._filter = kwargs
        return query

    def values(self, *args):
        query = self._chain()
        query._select = []
        for field in args:
            foreign_key, _, join_field = field.partition(".")
            if not join_field:
                query._select.append(field)
                continue
            if query._join is None:
                query._join = {}
            if foreign_key not in query._join:
                table = foreign_key[:-3]
                if table == "entity":
                    name, table = self.model._implicit_join
                else:
                    name = table.title()
                query._join[foreign_key] = (name, table)
            query._select.append(".".join((query._join[foreign_key][1], join_field)))
        return query

    def order_by(self, **kwargs):
        query = self._chain()
        query._order = kwargs
        return query

    def create(self, **kwargs):
        query = self._interface().values(kwargs)
        response = self._interface().execute("create", self._entity, query)
        return response

    def save(self, records: list[dict[str, str]], defaults: dict[str, str] | None = None, match: list[str] | None = None):
        query = {"records": records}
        if defaults is not None:
            query["defaults"] = defaults
        if match is not None:
            query["match"] = match
        response = self._interface().execute("save", self._entity, query)
        return response

    def delete(self):
        query = self._compose()
        response = self._interface().execute("delete", self._entity, query)
        return response

    def update(self, **kwargs):
        query = self._compose(kwargs)
        response = self._interface().execute("update", self._entity, query)
        return response



from civipy.base.base import CiviCRMBase


class CiviOrder(CiviCRMBase):
    pass


class CiviPayment(CiviCRMBase):
    @classmethod
    def cancel(cls, **kwargs):
        return cls.action("cancel", **kwargs)

    @classmethod
    def find_by_transaction_id(cls, trxn_id: str, select: list[str] | None = None) -> "CiviPayment | None":
        """Find a Contribution Payment by payment transaction ID"""
        found = cls.find_all(select=select, trxn_id=trxn_id)
        return next(filter(lambda c: bool(c.civi.get("contribution_id")), found), None)



from civipy.base.config import SETTINGS
from civipy.exceptions import CiviProgrammingError
from civipy.interface.base import CiviValue, CiviResponse, BaseInterface as Interface


def get_interface():
    if SETTINGS.api_version == "4":
        from civipy.interface.v4 import v4_interface

        return v4_interface
    if SETTINGS.api_version == "3":
        from civipy.interface.v3 import v3_interface

        return v3_interface
    raise CiviProgrammingError(f"Invalid API version '{SETTINGS.api_version}' - cannot initialize interface.")


__all__ = ["get_interface", "CiviValue", "CiviResponse", "Interface"]



from civipy.base.base import CiviCRMBase
from civipy.exceptions import NoResultError, NonUniqueResultError


class CiviUFField(CiviCRMBase): ...


class CiviUFGroup(CiviCRMBase): ...


class CiviUFJoin(CiviCRMBase): ...


class CiviUFMatch(CiviCRMBase):
    """This is the table that matches host system users to CiviCRM Contacts.

    create requires uf_id, uf_name, and contact_id

    Attributes:
        id: str e.g. "24392"
        domain_id: str e.g. "1"
        uf_id: str e.g. "46914"
        uf_name: str e.g. "user@example.com"
        contact_id: str e.g. "367872"
    """

    @classmethod
    def find_system_users(cls, contact_ids: list[int]) -> list["CiviUFMatch"]:
        result = []
        for contact_id in set(contact_ids):
            found = cls.find(contact_id=contact_id)
            if not found:
                continue
            result.append(found)
        if not result:
            raise NoResultError("No result found!")
        if len(result) != 1:
            views = [f"{r.civi}\n" for r in result]
            raise NonUniqueResultError(f"Too many results:\n {''.join(views)}")
        result = result[0]
        for attr in ("id", "domain_id", "uf_id", "contact_id"):
            result.civi[attr] = int(result.civi[attr])
        return result

    def update_system_user(self, user_id: int):
        return self.update(**self.civi, uf_id=user_id)

    @classmethod
    def connect(cls, host_user: int, contact_id: int, domain_id: int = 1):
        return cls.objects.create(domain_id=domain_id, uf_id=host_user, contact_id=contact_id)


class CiviUser(CiviCRMBase): ...



from civipy.base.base import CiviCRMBase


class CiviEntityFinancialTrxn(CiviCRMBase):
    @classmethod
    def cancel(cls, **kwargs):
        return cls.action("cancel", **kwargs)

    @classmethod
    def find_by_transaction_id(cls, trxn_id: str, entity_table: str) -> "CiviEntityFinancialTrxn | None":
        """Find a Contribution Payment by payment transaction ID"""
        found = cls.objects.filter(
            entity_table=entity_table, financial_trxn_id__trxn_id=trxn_id
        ).select(["*", "financial_trxn_id.*"]).all()
        return next(filter(lambda c: bool(c.civi.get("entity_id")), found), None)


class CiviFinancialTrxn(CiviCRMBase):
    ...


class CiviFinancialItem(CiviCRMBase):
    ...


class CiviLineItem(CiviCRMBase):
    ...



import json
from typing import TypeVar
from warnings import warn
from civipy.base.config import logger
from civipy.base.utils import get_unique
from civipy.exceptions import CiviProgrammingError
from civipy.interface import CiviValue, CiviResponse
from civipy.interface.query import Query

CiviEntity = TypeVar("CiviEntity", bound="CiviCRMBase")


class MetaCiviCRM(type):
    def __new__(typ, name, bases, dct):
        cls = super().__new__(typ, name, bases, dct)
        cls.objects = Query(cls)
        return cls


class CiviCRMBase(metaclass=MetaCiviCRM):
    objects: Query

    @classmethod
    def get(cls, **kwargs) -> CiviResponse:
        """Make an API request with the "get" action and return the full response."""
        warn("model.get will be removed in v0.1.0, use model.objects methods", DeprecationWarning, stacklevel=2)
        query = cls.objects._interface().limit(25)
        query.update(kwargs)
        return cls.action("get", **query)

    @classmethod
    def create(cls, **kwargs: CiviValue) -> CiviEntity:
        """Make an API request with the "create" action and return an object of class cls
        populated with the created object's data."""
        warn("model.create will be removed in v0.1.0, use model.save", DeprecationWarning, stacklevel=2)
        query = cls.objects._interface().values(kwargs)
        response = cls.action("create", **query)
        logger.debug("new record created! full response: %s" % str(response))
        return cls(get_unique(response))

    def update(self: CiviEntity, **kwargs: CiviValue) -> CiviResponse:
        """Update the current object with the values specified in kwargs. Returns the full
        API response."""
        warn("model.update will be removed in v0.1.0, use model.save", DeprecationWarning, stacklevel=2)
        kwargs["id"] = self.civi_id
        query = self.objects._interface().values(kwargs)
        return self.action("update", **query)

    def save(self: CiviEntity) -> CiviEntity:
        """Save the current instance."""
        action = "create" if self.civi_id is None else "update"
        query = self.objects._interface().values(self.civi)
        response = self.action(action, **query)
        return self.__class__(get_unique(response))

    def delete(self: CiviEntity, check_permissions: bool = True, use_trash: bool = True) -> CiviResponse:
        query = self.objects._interface().where({"id": self.civi_id})
        if check_permissions is False:
            query["checkPermissions"] = False
        if use_trash is False:
            query["useTrash"] = False
        return self.action("delete", **query)

    @classmethod
    def find(cls, select: list[str] | None = None, **kwargs: CiviValue) -> CiviEntity | None:
        """Looks for an existing object in CiviCRM with parameters equal to the values
        specified in kwargs. If using API v4 and select is specified, the result will include
        the specified keys.

        Returns an object of class cls populated with this object's data if found, otherwise
        returns None."""
        warn("model.find will be removed in v0.1.0, use model.objects methods", DeprecationWarning, stacklevel=2)
        query = cls.objects._interface().where(kwargs)
        if select:
            query["select"] = select
        response = cls.get(**query)
        if response["count"] == 0:
            return None
        return cls(get_unique(response))

    @classmethod
    def find_all(cls, select: list[str] | None = None, **kwargs: CiviValue) -> list[CiviEntity]:
        """Looks for multiple existing objects in CiviCRM with parameters equal to the
        values specified in kwargs. If using API v4 and select is specified, the result will
        include the specified keys.

        Returns a list of objects of class cls populated with data. Returns an empty list
        if no matching values found."""
        warn("model.find_all will be removed in v0.1.0, use model.objects methods", DeprecationWarning, stacklevel=2)
        query = cls.objects._interface().where(kwargs)
        if select:
            query["select"] = select
        response = cls.action("get", **query)
        return [cls(v) for v in response["values"]]

    @classmethod
    def find_and_update(cls, where: CiviValue, **kwargs: CiviValue) -> CiviEntity | None:
        """Looks for an existing object in CiviCRM with parameters equal to the values
        specified in `where`.

        If a unique record is found, record is also updated with values in `kwargs`.

        Returns an object of class cls populated with this object's data if found, otherwise
        returns None."""
        warn(
            "model.find_and_update will be removed in v0.1.0, use model.objects methods",
            DeprecationWarning,
            stacklevel=2,
        )
        query = cls.objects._interface().where(where)
        response = cls.get(**query)
        if response["count"] == 0:
            return None

        value = get_unique(response)
        value.update(kwargs)
        new_response = cls.action("create", **value)
        updated_value = get_unique(new_response)
        # not all fields are included in the return from an update, so we merge both sources
        updated_value.update(value)
        return cls(updated_value)

    @classmethod
    def find_or_create(cls, where: CiviValue, do_update: bool = False, **kwargs: CiviValue) -> CiviEntity:
        """Looks for an existing object in CiviCRM with parameters search_keys equal to the
        values for search_keys specified in kwargs.

        If a unique record is found and do_update is True, record is also updated with
        values in `kwargs`.

        If no record is found, a new record is created with the data in `where` and `kwargs`.

        Returns an object of class cls populated with the found, updated, or created
        object's data."""
        warn(
            "model.find_or_create will be removed in v0.1.0, use model.objects methods",
            DeprecationWarning,
            stacklevel=2,
        )
        if do_update:
            obj = cls.find_and_update(where, **kwargs)
        else:
            obj = cls.find(**where)

        if obj is None:
            query = where.copy()
            query.update(kwargs)
            return cls.objects.create(**query)
        return obj

    @classmethod
    def action(cls, action: str, **kwargs) -> CiviResponse:
        """Calls the CiviCRM API action and returns parsed JSON on success."""
        warn("model.action will be removed in v0.1.0, use model.objects methods", DeprecationWarning, stacklevel=2)
        entity = cls.__name__[4:]
        if entity == "CRMBase":
            raise CiviProgrammingError("Subclass CiviCRMBase to create an unsupported Entity.")
        return cls.objects._interface().execute(action, entity, kwargs)

    def pprint(self: CiviEntity) -> None:
        """Print the current record's data in a human-friendly format."""
        print(json.dumps(self.civi, sort_keys=True, indent=4))

    def __init__(self: CiviEntity, data: CiviValue) -> None:
        self.civi = data

    REPR_FIELDS = ["display_name", "name"]

    def __repr__(self: CiviEntity):
        label = None

        for field_name in self.REPR_FIELDS:
            if field_name in self.civi:
                label = self.civi[field_name]
                break

        return f"<{self.__class__.__name__} {self.civi_id}: {label}>"

    def __getattr__(self: CiviEntity, key: str):
        if key in self.civi:
            return self.civi[key]
        elif key.startswith("civi_") and key[5:] in self.civi:
            return self.civi[key[5:]]
        return None

    def __setattr__(self: CiviEntity, key: str, value: str | int | None) -> None:
        if key == "civi":
            object.__setattr__(self, key, value)
        elif key in self.civi:
            self.civi[key] = value
        elif key.startswith("civi_"):
            adj_key = key[5:]
            self.civi[adj_key] = value
        else:
            object.__setattr__(self, key, value)



import json
import subprocess
import urllib3
from civipy.base.config import SETTINGS, logger
from civipy.exceptions import CiviAPIError, CiviHTTPError, CiviProgrammingError
from civipy.interface.base import CiviValue, CiviV3Response, BaseInterface


class V3Interface(BaseInterface):
    __doc__ = (
        BaseInterface.__doc__
        + """
This is the v3 API interface."""
    )

    api_version = "3"

    def execute(self, action: str, entity: str, params: CiviValue) -> CiviV3Response:
        if self.func is None:
            if SETTINGS.api_version != "3":
                raise CiviProgrammingError(f"API version '{SETTINGS.api_version}' cannot use V3Interface")
            if SETTINGS.api_type == "http":
                self.func = self.http_request
                self._configure_http_connection()
            elif SETTINGS.api_type in ("drush", "wp"):
                self.func = self.run_drush_or_wp_process
            elif SETTINGS.api_type == "cvcli":
                self.func = self.run_cv_cli_process
            else:
                raise CiviProgrammingError(f"API type '{SETTINGS.api_type}' not implemented")
        # in v3, the 'update' action is deprecated and we are instructed to use 'create' with an id.
        if action == "update" and "id" in params:
            action = "create"
        return self.func(action, entity, params)

    def http_request(self, action: str, entity: str, kwargs: CiviValue) -> CiviV3Response:
        # v3 see https://docs.civicrm.org/dev/en/latest/api/v3/rest/
        kwargs = self._pre_process(kwargs)
        params = self._http_params(entity, action, kwargs)

        # header for v3 API per https://docs.civicrm.org/dev/en/latest/api/v3/rest/#x-requested-with
        headers = {"X-Requested-With": "XMLHttpRequest"}

        # v3 API GET actions apparently all start with "get"; POST actions are create, delete, etc.
        method = "GET" if action.startswith("get") else "POST"
        # urllib3 uses the `fields` parameter to compose the query string for GET requests,
        # and uses the same parameter to compose form data for POST requests
        response = self.http.request(method, SETTINGS.rest_base, fields=params, headers=headers)
        return self.process_http_response(response)

    @staticmethod
    def _http_params(entity: str, action: str, kwargs: CiviValue) -> CiviValue:
        params = {
            "entity": entity,
            "action": action,
            "api_key": SETTINGS.user_key,
            "debug": 1,
            "key": SETTINGS.site_key,
            "json": 1,
            "sequential": 1,
        }
        if kwargs:
            for key, value in kwargs.items():
                if value is None:
                    continue
                params[key] = value
        return params

    def process_http_response(self, response: urllib3.BaseHTTPResponse) -> CiviV3Response:
        logger.debug(response.url)
        if response.status == 200:
            return self.process_json_response(response.json())
        else:
            raise CiviHTTPError(response)

    def run_cv_cli_process(self, action: str, entity: str, params: CiviValue) -> CiviV3Response:
        params = self._pre_process(params)
        # cli.php -e entity -a action [-u user] [-s site] [--output|--json] [PARAMS]
        params = ["--%s=%s" % (k, v) for k, v in params.items()]
        process = subprocess.run(
            [SETTINGS.rest_base, "-e", entity, "-a", action, "--json"] + params, capture_output=True, check=True
        )
        return self.process_json_response(json.loads(process.stdout.decode("UTF-8")))

    def run_drush_or_wp_process(self, action: str, entity: str, params: CiviValue) -> CiviV3Response:
        params = self._pre_process(params)
        process = subprocess.run(
            [SETTINGS.rest_base, "civicrm-api", "--out=json", "--in=json", "%s.%s" % (entity, action)],
            capture_output=True,
            check=True,
            input=json.dumps(params).encode("UTF-8"),
        )
        return self.process_json_response(json.loads(process.stdout.decode("UTF-8")))

    @staticmethod
    def _pre_process(params: CiviValue) -> CiviValue:
        if "options" in params:
            params["options"] = json.dumps(params["options"], separators=(",", ":"))
        return params

    @staticmethod
    def process_json_response(data: CiviV3Response) -> CiviV3Response:
        if "is_error" in data and data["is_error"] == 1:
            raise CiviAPIError(data)
        return data

    @staticmethod
    def select(fields: list[str]) -> CiviValue:
        return {"return": fields}

    @staticmethod
    def join(tables: list[tuple[str, str, str]]) -> CiviValue:
        raise NotImplementedError

    @staticmethod
    def sort(kwargs: CiviValue) -> CiviValue:
        option = []
        for k, v in kwargs.items():
            if isinstance(v, str) and v.upper() in ("ASC", "DESC"):
                option.append(k if v.upper() == "ASC" else f"{k} DESC")
            elif isinstance(v, int) and v in (0, 1):
                option.append(k if v else f"{k} DESC")
            else:
                raise CiviProgrammingError(f"Invalid sort value for {k}: {repr(v)}")
        return {"options": {"sort": option}}

    @staticmethod
    def limit(value: int, offset: int | None = None) -> CiviValue:
        option = {"limit": value}
        if offset is not None:
            option["offset"] = offset
        return {"options": option}

    @classmethod
    def where(cls, kwargs: CiviValue) -> CiviValue:
        option = {}
        for key, val in kwargs.items():
            parts = key.split("__")
            if len(parts) > 1 and parts[-1] in cls.operators:
                *parts, op = parts
                if op == "isnull":
                    val = {"IS NULL": 1} if val else {"IS NOT NULL": 1}  # noqa PLW2901 (loop var)
                elif (op.endswith("between") or op.endswith("in")) and not isinstance(val, list):
                    raise CiviProgrammingError(f"Must provide a list for `{op}` operator.")
                else:
                    val = {cls.operators[op]: val}  # noqa PLW2901 (loop var)
            option[".".join(parts)] = val
        return kwargs

    @staticmethod
    def values(kwargs: CiviValue) -> CiviValue:
        return kwargs


v3_interface = V3Interface()
__all__ = ["v3_interface"]



import json
import subprocess
from urllib.parse import urljoin, urlencode
import urllib3
from civipy.base.config import SETTINGS, logger
from civipy.exceptions import CiviAPIError, CiviHTTPError, CiviProgrammingError
from civipy.interface.base import CiviValue, CiviV4Response, BaseInterface, CiviV4Request


class V4Interface(BaseInterface):
    __doc__ = (
        BaseInterface.__doc__
        + """
This is the v4 API interface."""
    )

    api_version = "4"

    def execute(self, action: str, entity: str, params: CiviValue) -> CiviV4Response:
        if self.func is None:
            if SETTINGS.api_version != "4":
                raise CiviProgrammingError(f"API version '{SETTINGS.api_version}' cannot use V4Interface")
            if SETTINGS.api_type == "http":
                self.func = self.http_request
                self._configure_http_connection()
            elif SETTINGS.api_type == "drush":
                self.func = self.run_drush_process
            # API v4 not available to wp-cli - see https://docs.civicrm.org/dev/en/latest/api/v4/usage/#wp-cli
            elif SETTINGS.api_type == "cvcli":
                self.func = self.run_cv_cli_process
            else:
                raise CiviProgrammingError(f"API type '{SETTINGS.api_type}' not implemented")
        return self.func(action, entity, params)

    def http_request(self, action: str, entity: str, kwargs: CiviV4Request) -> CiviV4Response:
        # v4 see https://docs.civicrm.org/dev/en/latest/api/v4/rest/
        url = urljoin(SETTINGS.rest_base, "/".join((entity, action)))
        params = json.dumps(kwargs, separators=(",", ":"))
        body = urlencode({"params": params})
        logger.debug("Request for %s: %s", url, params)

        # header for v4 API per https://docs.civicrm.org/dev/en/latest/api/v4/rest/#x-requested-with
        headers = {
            "X-Requested-With": "XMLHttpRequest",
            "X-Civi-Auth": f"Bearer {SETTINGS.user_key}",
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # v4 docs: "Requests are typically submitted with HTTP POST, but read-only operations may use HTTP GET."
        response = self.http.request("POST", url, body=body, headers=headers)
        return self.process_http_response(response)

    def process_http_response(self, response: urllib3.BaseHTTPResponse) -> CiviV4Response:
        logger.debug(response.url)
        if response.status == 200:
            return self.process_json_response(response.json())
        else:
            raise CiviHTTPError(response)

    def run_cv_cli_process(self, action: str, entity: str, params: CiviValue) -> CiviV4Response:
        # see `cv --help api4` or https://docs.civicrm.org/dev/en/latest/api/v4/usage/#cv
        process = subprocess.run(
            [
                SETTINGS.rest_base,
                "api4",
                ".".join((entity, action)),
                json.dumps(params, separators=(",", ":")).encode("UTF-8"),
            ],
            capture_output=True,
            check=True,
        )
        return self.process_json_response(json.loads(process.stdout.decode("UTF-8")))

    def run_drush_process(self, action: str, entity: str, params: CiviValue) -> CiviV4Response:
        process = subprocess.run(
            [SETTINGS.rest_base, "civicrm-api", "version=4", "--out=json", "--in=json", "%s.%s" % (entity, action)],
            capture_output=True,
            check=True,
            input=json.dumps(params, separators=(",", ":")).encode("UTF-8"),
        )
        return self.process_json_response(json.loads(process.stdout.decode("UTF-8")))

    @staticmethod
    def process_json_response(data: CiviV4Response) -> CiviV4Response:
        if "is_error" in data and data["error_code"] > 0:
            raise CiviAPIError(data)
        return data

    @staticmethod
    def select(fields: list[str]) -> CiviV4Request:
        return {"select": fields}

    @staticmethod
    def join(tables: dict[str, tuple[str, str]]) -> CiviV4Request:
        option = []
        for foreign_key, (name, table) in tables.items():
            option.append([f"{name} AS {table}", "LEFT", [foreign_key, "=", f"{table}.id"]])
        return {"join": option}

    @staticmethod
    def sort(kwargs: CiviValue) -> CiviValue | CiviV4Request:
        option = {}
        for k, v in kwargs.items():
            if isinstance(v, str) and v.upper() in ("ASC", "DESC"):
                option[k] = v.upper()
            elif isinstance(v, int) and v in (0, 1):
                option[k] = ("DESC", "ASC")[v]
            else:
                raise CiviProgrammingError(f"Invalid sort value for {k}: {repr(v)}")
        return {"orderBy": option}

    @staticmethod
    def limit(value: int, offset: int | None = None) -> CiviV4Request:
        option = {"limit": value}
        if offset is not None:
            option["offset"] = offset
        return option

    operators = BaseInterface.operators | {
        "contains": "CONTAINS",
        "not_contains": "NOT CONTAINS",
        "isempty": "IS EMPTY",
        "regexp": "REGEXP",
        "not_regexp": "NOT REGEXP",
    }

    @classmethod
    def where(cls, kwargs: CiviValue) -> CiviV4Request:
        option = []
        for key, val in kwargs.items():
            parts = key.split("__")
            if len(parts) > 1 and parts[-1] in cls.operators:
                *parts, op = parts
                if op == "isnull":
                    value = ["IS NULL"] if val else ["IS NOT NULL"]
                elif op == "isempty":
                    value = ["IS EMPTY"] if val else ["IS NOT EMPTY"]
                elif (op.endswith("between") or op.endswith("in")) and not isinstance(val, list):
                    raise CiviProgrammingError(f"Must provide a list for `{op}` operator.")
                else:
                    value = [cls.operators[op], val]
            else:
                value = ["=", val]
            option.append([".".join(parts)] + value)
        return {"where": option}

    @staticmethod
    def values(kwargs: CiviValue) -> CiviV4Request:
        return {"values": kwargs}


v4_interface = V4Interface()
__all__ = ["v4_interface"]



from civipy.base.base import CiviCRMBase


class CiviOptionValue(CiviCRMBase):
    pass


class CiviOptionGroup(CiviCRMBase):
    @classmethod
    def find_options_by_group_name(cls, option_group_name: str):
        """
        Taking an option_group_name, looks up the group and its members.
        """
        og = CiviOptionGroup.find(name=option_group_name)
        return CiviOptionValue.find_all(option_group_id=og["id"])

    @classmethod
    def option_values_dict_by_group_name(cls, option_group_name):
        """
        Taking an option_group_name, looks up the group and its members.
        """
        option_values_list = cls.find_options_by_group_name(option_group_name)
        if isinstance(option_values_list, dict):
            option_values_list = option_values_list.values()
        return dict((t["name"], t["value"]) for t in option_values_list)


class CiviCustomValue(CiviCRMBase):
    pass


class CiviCustomField(CiviCRMBase):
    @classmethod
    def find_field_by_label(cls, field_label: str):
        return CiviCustomField.find(label=field_label)

    @classmethod
    def find_options_by_field_label(cls, label: str):
        custom_field = cls.find(label=label)
        return CiviOptionValue.find_all(option_group_id=custom_field.civi_option_group_id)

    @classmethod
    def options_label_map(cls, label: str):
        option_values = cls.find_options_by_field_label(label)
        return dict((ov["label"], ov["value"]) for ov in option_values)
